/*
 * Copyright (c) 2019-2021, NVIDIA CORPORATION.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

#include <limits>

#include <cuda/std/chrono>

/**
 * @file timestamps.hpp
 * @brief Concrete type definitions for int32_t and int64_t timestamps in
 * varying resolutions as durations since the UNIX epoch.
 */
namespace cudf {
namespace detail {
// TODO: Use chrono::utc_clock when available in libcu++?
template <class Duration>
using time_point = cuda::std::chrono::sys_time<Duration>;

template <class Duration>
using timestamp = time_point<Duration>;
}  // namespace detail

/**
 * @addtogroup timestamp_classes
 * @{
 * @file
 */

/**
 * @brief Type alias representing an int32_t duration of days since the unix
 * epoch.
 */
using timestamp_D =
  detail::timestamp<cuda::std::chrono::duration<int32_t, cuda::std::ratio<86400>>>;
/**
 * @brief Type alias representing an int64_t duration of seconds since the
 * unix epoch.
 */
using timestamp_s = detail::timestamp<cuda::std::chrono::duration<int64_t, cuda::std::ratio<1>>>;
/**
 * @brief Type alias representing an int64_t duration of milliseconds since
 * the unix epoch.
 */
using timestamp_ms = detail::timestamp<cuda::std::chrono::duration<int64_t, cuda::std::milli>>;
/**
 * @brief Type alias representing an int64_t duration of microseconds since
 * the unix epoch.
 */
using timestamp_us = detail::timestamp<cuda::std::chrono::duration<int64_t, cuda::std::micro>>;
/**
 * @brief Type alias representing an int64_t duration of nanoseconds since
 * the unix epoch.
 */
using timestamp_ns = detail::timestamp<cuda::std::chrono::duration<int64_t, cuda::std::nano>>;

static_assert(sizeof(timestamp_D) == sizeof(typename timestamp_D::rep), "");
static_assert(sizeof(timestamp_s) == sizeof(typename timestamp_s::rep), "");
static_assert(sizeof(timestamp_ms) == sizeof(typename timestamp_ms::rep), "");
static_assert(sizeof(timestamp_us) == sizeof(typename timestamp_us::rep), "");
static_assert(sizeof(timestamp_ns) == sizeof(typename timestamp_ns::rep), "");

/** @} */  // end of group
}  // namespace cudf

namespace std {
/**
 * @brief Specialization of std::numeric_limits for cudf::detail::timestamp
 *
 * Pass through to return the limits of the underlying numeric representation.
 */
#define TIMESTAMP_LIMITS(TypeName)                                                                \
  template <>                                                                                     \
  struct numeric_limits<TypeName> {                                                               \
    static constexpr TypeName max() noexcept { return TypeName::max(); }                          \
    static constexpr TypeName lowest() noexcept                                                   \
    {                                                                                             \
      return TypeName{TypeName::duration{std::numeric_limits<typename TypeName::rep>::lowest()}}; \
    }                                                                                             \
    static constexpr TypeName min() noexcept { return TypeName::min(); }                          \
  }

TIMESTAMP_LIMITS(cudf::timestamp_D);
TIMESTAMP_LIMITS(cudf::timestamp_s);
TIMESTAMP_LIMITS(cudf::timestamp_ms);
TIMESTAMP_LIMITS(cudf::timestamp_us);
TIMESTAMP_LIMITS(cudf::timestamp_ns);

#undef TIMESTAMP_LIMITS

}  // namespace std
